/* Userstat - a simple front end to veiw SGE queue and host information
 * 
 *
 * Copyright 2005, 2011 Douglas Eadline deadline@basement-supercomputing.com
 * This code is licensed under the GNU Public General Public License
 * which is available from<F2> http://www.gnu.org/copyleft/gpl.html
 *  
 * AUTHOR:deadline
 * use "set ts=2" for best resutls
 * To build:  gcc -o  userstat src/userstat.c -lncurses
 *
 * quick ncures info
 *  http://www.apmaths.uwo.ca/~xli/ncurses.html
 *
 * Version 1.1 1/11/11
 * Only works with SGE 6.2
 * Basically we execlp qhost and qstat and pipe the output
 * into this program. We then parse the output and place it
 * into the data structures below. Of couse, a database may be better
 * if large amounts of reporting/sorting/loging  are required ....
 *
 * Works for sun grid engine only version 6.2u5
 * code needs  clean up and commenting. Parsing the qstat
 * output is the most convoluted then comes the window scrolling
 * stuff.
 * To fix or debug, use ddd graphical debugger and watch the
 * how the lines are parsed (queue data is the most messy)
 * compile with "-g" for debugging.
 *
 * Things to do:
 *  - check qstat -f for suspended and stoped queues (show in host window)
 *  - check for qstat and qhost binaries
 *  - possibly use gstat (ganglia instead of qhost)
 *  - support pbs
 *  - return key on a node produces detailed info (good for dual nodes)
 *    maybe open a remote top?
 *  - change the submit start date on running programs to 
 *    a duration?
 *
 * HISTORY: this code was intially hacked together while testing
 *          sge scripts. I tired of look at qstat and trying to figure
 *          out what the heck was going on. There is 
 *          probably a cleaner way to do with an sge api,
 *          but at the time, hacking a a convoluted C program
 *          seemed easier -- go figure. Plus,  main() -- 322 lines!
 *          You have been warned :-)
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <curses.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#define STRINGLEN 32
#define NODELISTLEN 1024
#define MAXHOSTS 1024
#define MAXJOBS 1024
#define MAXTOPDATA  8192
#define MAXLINELEN  2048
#define SKIP1 "HOSTNA"
#define SKIP2 "------"
#define SKIP3 "global"
#define SKIP4 "job-ID"

#define CLDEBUG  0
#define HOSTWINDOW 1
#define QUEUEWINDOW 2
#define RETRIES 10
 struct hostdata{
	char hostname[STRINGLEN];
	int processors;
	int down;
	float load;
	float mem;
	float memuse;
	float swap;
	float swapuse;
} hostdata;
struct queuedata{
	char id[STRINGLEN];
	char priority[STRINGLEN];
	char jobname[STRINGLEN];
	char username[STRINGLEN];
	char submittime[STRINGLEN];
	char submitdate[STRINGLEN];
	char state[STRINGLEN];
	int  nprocs;
	int  nodelist[NODELISTLEN];
}queuedata;

struct  hostdata hostlist[MAXHOSTS];
struct  queuedata joblist[MAXJOBS];
int nodecount=0,jobcount=0,tprocessors=0,jobindex=0;
int active,down,jobsortcount,jobnodecount, injobselect=0, injobsort=0;
int qhost_found,qmaster_up, top_found;
float tload=0,tmem=0,tmemuse=0,tswap=0,tswapuse=0;
char top_node[STRINGLEN]; 

void display_help();
void display_top();
void gethostdata();
void getqueuedata();
void gettopdata(char *top_data,int max_line_len, int max_lines);
void writehostwindow(int row,int hostwinheight,int start,int cursor,int cw,char *job, char *sortnodename);
void writequeuewindow(int row,int hostwinheight,int start,int cursor, int cw,char *job, char *sortusername,char *sortjobname,char *sortprogramname);
void do_text_error(char *error);
float mem_convert(char *value);
void print_totals(int y,int x);
char *pipefgets(char *line,size_t count,int fd);
pid_t pipe2qhost(int fdr,int fdw);
pid_t pipe2qstat(int fdr, int fdw);
pid_t pipe2top(int fdr, int fdw);
void delete_job(char *jobname);

#define FD_READ 0
#define FD_WRITE 1
#define HOSTLABLE2 "Host          CPUs     Load     Memory     Memory Use    Swap    Swap Use"
#define QUEUELABLE2 " Job-ID  Prior   Name            User   State     Submit/Start       CPUs"
#define HELPBARSTRING "h-Help   q-Quit   b-Batch Window   n-Node Window   spacebar-Update   v1.1 "
#define HELPBARSTRCLR "                                                                   "
#define REFRESHINTERVAL 20 
#define NOQSTATSTRING "ERROR: qstat and qhost not found or qmaster not running.  Enter q to quit."
main(void){
	
	int maxx,maxy,statline,queuelable,hostlable,helpbar,middle;
	int queuelable2,hostlable2,i,key,delindex;
	int queuewinheight,queuewinwidth,hostwinheight,hostwinwidth;
	int selection=0, currentwindow, currentqcursor=0, currenthcursor=0;
	int qstartline=0, hstartline=0, tempjobcount, tempnodecount;
	time_t lastrefresh, timenow, dummy;
	WINDOW *queuewin, *hostwin;
	FILE *hostinfo,*queueinfo;
	char filename[STRINGLEN], jobnodes[STRINGLEN], jobonly[STRINGLEN],sortusername[STRINGLEN];
	char sortjobname[STRINGLEN],sortprogramname[STRINGLEN],sortnodename[STRINGLEN];
	struct stat fileinfo;
	/* collect data from information files, place in queuedata 
   * and hostdata structures for display. 
   */
	currentwindow=QUEUEWINDOW;
	
#if !CLDEBUG
	strcpy(jobonly,"");
	strcpy(sortusername,"");
	strcpy(sortprogramname,"");
	strcpy(sortjobname,"");
	strcpy(sortnodename,"");
	initscr();
	cbreak();
	keypad(stdscr,TRUE);
	getmaxyx(stdscr,maxy,maxx);
	if(maxy<10) do_text_error("window height too small");
	if(maxx<strlen(HOSTLABLE2)) do_text_error("window width too small");
	gethostdata();
	getqueuedata();
	do {

		clear();
		getmaxyx(stdscr,maxy,maxx);
		/* calculate/assign positions */
		queuelable=0;
		queuelable2=queuelable+1;
		hostlable=maxy/2;
		hostlable2=hostlable+1;
		statline=hostlable+2;
		helpbar=maxy-1;
		middle=maxx/2;
		queuewinheight=hostlable-(queuelable+1)-1;
		queuewinwidth=maxx;
		hostwinheight=helpbar-(statline+1);
		hostwinwidth=maxx;
		attron(A_REVERSE);
		move(hostlable2,0);
    printw(HOSTLABLE2);
		move(queuelable2,0);
		printw(QUEUELABLE2);
		move(helpbar,0);
		if (qhost_found==0) printw(NOQSTATSTRING);
		else printw(HELPBARSTRING);
		attroff(A_REVERSE);
		/*printw("currenthcursor=%i,hstartline=%i hostwinheight%i jobnodecount=%i nodecount=%i\n",
			currenthcursor,hstartline,hostwinheight,jobnodecount,nodecount);
		printw("curqcursor=%i,qstrtline=%i queuewinhght%i jobcount=%i nodecount=%i jobsortcnt=%i\n",
			currenthcursor,hstartline,hostwinheight,jobcount,nodecount,jobsortcount);*/
		if(currentwindow==HOSTWINDOW) attron(A_REVERSE);
		move(hostlable,0);
		attron(A_BOLD);
		if(injobselect&&(inarray(jobonly,jobcount))!=-1)
			printw("HOSTS Total Nodes: %i  Down nodes: %i  Nodes for Job-ID: %s",nodecount, down,jobonly);
		else
			printw("HOSTS Total Nodes: %i  Down nodes: %i",nodecount, down);
		attroff(A_BOLD);
		if(currentwindow==HOSTWINDOW) attroff(A_REVERSE);
		attroff(A_REVERSE);
		print_totals(statline,0);
		writequeuewindow(queuelable2+1,queuewinheight,qstartline,currentqcursor,
			currentwindow,jobnodes,sortusername,sortjobname,sortprogramname);
		writehostwindow(hostlable2+2,hostwinheight,hstartline,currenthcursor,
			currentwindow,jobonly,sortnodename);
		if(currentwindow==QUEUEWINDOW) attron(A_REVERSE); /* do this after display because we count active jobs*/
		move(queuelable,0);
		attron(A_BOLD);
		printw("BATCH QUEUE Total Jobs: %i   Active Jobs: %i",jobcount,active);
		attroff(A_BOLD);
		if(currentwindow==QUEUEWINDOW) attroff(A_REVERSE);
		/*move(queuelable,0);
		printw("curqcursor=%i,qstrtline=%i queuewinhght%i jobcount=%i \n",
			currentqcursor,qstartline,queuewinheight,jobcount);
		*/
		
		if (currentwindow==QUEUEWINDOW) move(qstartline,0);
		else move(hstartline,0);
		/*mark the scroll "bars"*/
		if(currentwindow==QUEUEWINDOW){
			if(injobsort) tempjobcount=jobsortcount;
			else tempjobcount=jobcount;
			if(qstartline>0) {
				move(queuelable2+1,0);printw("+");
			}
			if(qstartline+queuewinheight<tempjobcount){
				move(queuelable2+queuewinheight,0);printw("+");
			}
			move(queuelable2+1+currentqcursor-qstartline,0);
		}
		else{
			if(injobselect) tempnodecount=jobnodecount;
			else tempnodecount=nodecount;
			if(hstartline>0) {
				move(hostlable2+2,0);printw("+");
			}
			if(hstartline+hostwinheight<tempnodecount){
				move(hostlable2+1+hostwinheight,0);printw("+");
			}
			move(hostlable2+2+currenthcursor-hstartline,0);
		}

		refresh();  /*THIS IS WHERE THE WINDOW GETS WRITTEN */
		lastrefresh=time(&dummy);
  	noecho();
		/*nodelay(stdscr, TRUE);*/
		cbreak();
		halfdelay(5);
	  while((selection = getch()) == ERR){
	 		timenow=time(&dummy);
			/*sleep(1);*/
			if ((timenow-lastrefresh)>REFRESHINTERVAL){
				selection=' ';
				break;
			}
		}
		nocbreak();
		/*nodelay(stdscr, FALSE);*/
  	echo();
			move(helpbar,0);
		switch(selection){
			case 'd':
				if(currentwindow==QUEUEWINDOW){
					if((delindex=inarray(jobnodes,jobcount))!=-1) {
						move(helpbar,0);
						printw(HELPBARSTRCLR);
						noecho();
						cbreak();
						while(1){
							move(helpbar,0);
							printw("Delete Job %s (y/n)? ",joblist[delindex].id);
							key=getch();
							if (key=='y'){
								delete_job(joblist[delindex].id);
								break;
							}else if (key=='n') break;
						}
						nocbreak();
						echo();
						strcpy(jobonly,"");
						strcpy(sortnodename,"");
						injobselect=0;
						hstartline=0;
						currenthcursor=0;				
					}
			}
			break;
			case 'q':
				endwin();
				exit(0);
			case ' ':
				tprocessors=0;
				tload=0;
				tmem=0;
				tmemuse=0;
				tswap=0;
				tswapuse=0;
				jobcount=0;
				nodecount=0;
				gethostdata();
				getqueuedata();
				break;	
			case 'h':
				move(0,0);
				display_help();
				break;	
			case 'n':
				currentwindow=HOSTWINDOW;
				break;	
			case 'j':
				if(currentwindow==QUEUEWINDOW){
					move(helpbar,0);
					printw(HELPBARSTRCLR);
					move(helpbar,0);
					printw("Enter Job-ID Number:  ");
					i=0;
					while ((key=getch())!='\n'){
						sortjobname[i++]=key;
						if (i>=STRINGLEN) break;
					}
					sortjobname[i]=0;	
					qstartline=0;
					currentqcursor=0;
					break;	
				}
			case 'p':
				if(currentwindow==QUEUEWINDOW){
					move(helpbar,0);
					printw(HELPBARSTRCLR);
					move(helpbar,0);
					printw("Enter Program Name:  ");
					i=0;
					while ((key=getch())!='\n'){
						sortprogramname[i++]=key;
						if (i>=STRINGLEN) break;
					}
					sortprogramname[i]=0;	
					hstartline=0;
					currenthcursor=0;
				}
				break;	
			case 's':  /*only if hostwindow and we have not sorted by job*/
				if(currentwindow==HOSTWINDOW&&injobselect==0){
					move(helpbar,0);
					printw(HELPBARSTRCLR);
					move(helpbar,0);
					printw("Enter System Name:  ");
					i=0;
					while ((key=getch())!='\n'){
						sortnodename[i++]=key;
						if (i>=STRINGLEN) break;
					}
					sortnodename[i]=0;	
					hstartline=0;
					currenthcursor=0;
				}
				break;	
			case 'u':
				if(currentwindow==QUEUEWINDOW){
					move(helpbar,0);
					printw(HELPBARSTRCLR);
					move(helpbar,0);
					printw("Enter User Name:  ");
					i=0;
					while ((key=getch())!='\n'){
						sortusername[i++]=key;
						if (i>=STRINGLEN) break;
					}
					sortusername[i]=0;	
					currentwindow=QUEUEWINDOW;
					hstartline=0;
					currenthcursor=0;
				}
				break;	
			case 'b':
				currentwindow=QUEUEWINDOW;
				break;	
			case '\n':
				if(currentwindow==QUEUEWINDOW){
					injobselect=1;
					strcpy(jobonly,jobnodes);
					hstartline=0;
					currenthcursor=0;				
				}	
        /* show top data for node*/					
				else{
					 move(0,0);
        	display_top();
				}
				break;
			case 'a':
				if(currentwindow==HOSTWINDOW){
					strcpy(jobonly,"");
					strcpy(sortnodename,"");
					injobselect=0;
					hstartline=0;
					currenthcursor=0;				
				}
				else{
					strcpy(sortusername,"");
					strcpy(sortjobname,"");
					strcpy(sortprogramname,"");
				}
				break;	
			case KEY_UP: /* need to decrement the display start value as well*/
				if(currentwindow==QUEUEWINDOW){
					if(currentqcursor!= 0){
						 if(currentqcursor==qstartline) qstartline--;
						 currentqcursor--;
					}
					else
						if(qstartline!=0) qstartline--;
				}
				else{
					if(currenthcursor!=0){
						 if(currenthcursor==hstartline) hstartline--;
						 currenthcursor--;
					}
					else
						if(hstartline!=0) hstartline--;
				}	
				break;
			case KEY_DOWN: /* need to increment the display start as well */
				if(currentwindow==QUEUEWINDOW){
					if(injobsort){
							if(jobsortcount==0) tempjobcount=1; /* when no jobs to display*/
							else tempjobcount=jobsortcount;    /*  display job nodes only*/
					}
					else tempjobcount=jobcount;
					if((currentqcursor-qstartline)<queuewinheight-1&&currentqcursor!=tempjobcount-1)
						 currentqcursor++;
					else{
						if(qstartline!=tempjobcount-1&&currentqcursor!=tempjobcount-1){
							qstartline++;
							currentqcursor++;
						}
					}
				}
				else{
					if(injobselect){
							if(jobnodecount==0) tempnodecount=1; /* when no host to display*/
							else tempnodecount=jobnodecount;   /*display job host nodes only*/
					}
					else tempnodecount=nodecount;
					if((currenthcursor-hstartline)<hostwinheight-1&&currenthcursor!=tempnodecount-1)
						 currenthcursor++;
					else{
						if(hstartline!=tempnodecount-1&&currenthcursor!=tempnodecount-1){
							hstartline++;
							currenthcursor++;
						}
					}
				}
				break;	
			case KEY_PPAGE: /* need to increment the display start as well */
				if(currentwindow==QUEUEWINDOW){
					if((qstartline-queuewinheight-1)>0){
						qstartline-=queuewinheight-1;
						currentqcursor=qstartline;
					}
					else{
						qstartline=0;
						currentqcursor=qstartline;
					}
				}
				else{
					if((hstartline-hostwinheight-1)>0){
						hstartline-=(hostwinheight-1);
						currenthcursor=hstartline;
					}
					else{
						hstartline=0;
						currenthcursor=hstartline;
					}
				}
				break;
			case KEY_NPAGE: /* need to increment the display start as well */
				if(currentwindow==QUEUEWINDOW){
					if(injobsort){
							if(jobsortcount==0) tempjobcount=1; /* when no jobs to display*/
							else tempjobcount=jobsortcount;    /*  display job nodes only*/
					}
					else tempjobcount=jobcount;
					if((qstartline+queuewinheight-1)<tempjobcount)
						qstartline+=(queuewinheight-1);
					if(currentqcursor<tempjobcount)
								 currentqcursor=qstartline;
				}
				else{
					if(injobselect){
							if(jobnodecount==0) tempnodecount=1; /* when no host to display*/
							else tempnodecount=jobnodecount;   /*display job host nodes only*/
					}
					else tempnodecount=nodecount;
					if((hstartline+hostwinheight-1)<tempnodecount)
						hstartline+=hostwinheight-1;
					if(currenthcursor<tempnodecount)
								 currenthcursor=hstartline;
				}
				break;	
		}
	}while (1);
#endif
}
void print_totals(int y,int x){
  int i,memsize,memusesize,swapsize,swapusesize;
  float ttmem,ttmemuse,ttswap,ttswapuse;
  
	ttmem=tmem;
  ttmemuse=tmemuse;
  ttswap=tswap;
  ttswapuse=tswapuse;
  memsize=mem_unit(&ttmem);
  memusesize=mem_unit(&ttmemuse);
  swapsize=mem_unit(&ttswap);
  swapusesize=mem_unit(&ttswapuse);
	move(y,x);
	attron(A_BOLD);
  printw("cluster       %4i    %5.2f  %8.1f%c   %8.1f%c  %8.1f%c   %8.1f%c\n",
		tprocessors,tload,ttmem,memsize,ttmemuse,memusesize,ttswap,swapsize,
		ttswapuse,swapusesize);
	attroff(A_BOLD);
}
void writehostwindow(int row,int hostwinheight,int start,int cursor, int cw,char *jobonly,char *sortnodename){

	int i,memsize,memusesize,swapsize,swapusesize,jobonlyindex,dtemp;
	float tmem,tmemuse,tswap,tswapuse;
	char temp[STRINGLEN];
	/*do not forget to reduce nodecount to number of nodes so windows work*/
	move(row,0);
	if((jobonlyindex=inarray(jobonly,jobcount))!=-1) {
		jobnodecount=joblist[jobonlyindex].nprocs;
  	for (i=start+1;i<=jobnodecount&&i<=(hostwinheight+start);i++){
			dtemp=joblist[jobonlyindex].nodelist[i];
			tmem=hostlist[joblist[jobonlyindex].nodelist[i]].mem;
			tmemuse=hostlist[joblist[jobonlyindex].nodelist[i]].memuse;
			tswap=hostlist[joblist[jobonlyindex].nodelist[i]].swap;
			tswapuse=hostlist[joblist[jobonlyindex].nodelist[i]].swapuse;
			memsize=mem_unit(&tmem);
			memusesize=mem_unit(&tmemuse);
			swapsize=mem_unit(&tswap);
			swapusesize=mem_unit(&tswapuse);
			if(hostlist[joblist[jobonlyindex].nodelist[i]].down){
				strcpy(temp,"*");
				strcat(temp,hostlist[joblist[jobonlyindex].nodelist[i]].hostname);
				strcat(temp,"*");
			}
			else{
				strcpy(temp,hostlist[joblist[jobonlyindex].nodelist[i]].hostname);
			}
			if(i==cursor+1&&cw==HOSTWINDOW) {
				strcpy(top_node,temp); /* copy this node name for use in top display */				
				attron(A_BOLD);
			}
     	printw("%12s  %4i    %5.2f  %8.1f%c   %8.1f%c  %8.1f%c   %8.1f%c\n",
				temp,
				hostlist[joblist[jobonlyindex].nodelist[i]].processors,
				hostlist[joblist[jobonlyindex].nodelist[i]].load,
				tmem,memsize,tmemuse,memusesize,tswap,swapsize,tswapuse,swapusesize);
			if(i==cursor+1&&cw==HOSTWINDOW) attroff(A_BOLD);
    }
	}
	else{	
  	for (i=start;i<nodecount&&i<(hostwinheight+start);i++){
			tmem=hostlist[i].mem;
			tmemuse=hostlist[i].memuse;
			tswap=hostlist[i].swap;
			tswapuse=hostlist[i].swapuse;
			memsize=mem_unit(&tmem);
			memusesize=mem_unit(&tmemuse);
			swapsize=mem_unit(&tswap);
			swapusesize=mem_unit(&tswapuse);
			if(hostlist[i].down){
				strcpy(temp,"*");
				strcat(temp,hostlist[i].hostname);
				strcat(temp,"*");
			}
			else{
				strcpy(temp,hostlist[i].hostname);
			}
			if(!strcmp(sortnodename,"")||!strcmp(sortnodename,hostlist[i].hostname)){
				if(i==cursor&&cw==HOSTWINDOW){
					strcpy(top_node,temp); /* copy this node name for use in top display */				
					attron(A_BOLD);
				}
     			printw("%12s  %4i    %5.2f  %8.1f%c   %8.1f%c  %8.1f%c   %8.1f%c\n",
					temp,
					hostlist[i].processors,
					hostlist[i].load,tmem,
					memsize,tmemuse,memusesize,tswap,swapsize,tswapuse,swapusesize);
					if(i==cursor&&cw==HOSTWINDOW) attroff(A_BOLD);
					if(strcmp(sortnodename,"")) {
					jobnodecount=1;
					injobselect=1;
				}
			}
    }
	}
}
void writequeuewindow(int row,int queuewinheight,int start, int cursor,int cw,char *job,char *sortusername,char *sortjobname,char *sortprogramname){
	int i, printed=0;
	/*move(0,0);printw("sortprogramname=%s",sortprogramname);*/
	move(row,0);
	active=0;
	for (i=0;i<jobcount;i++){
			if(!strcmp(joblist[i].state,"r") /* only r and t */
				||!strcmp(joblist[i].state,"t")
				||!strcmp(joblist[i].state,"Rt")
				||!strcmp(joblist[i].state,"dr")
				||!strcmp(joblist[i].state,"dt"))
					active++;
	}	
	for (i=start;i<jobcount&&i<(queuewinheight+start);i++){
			if(strcmp(sortusername,"")
					||strcmp(sortjobname,"")
					||strcmp(sortprogramname,"")){ /*we are asked to sort */
				injobsort=1;
				if(!strcmp(sortusername,joblist[i].username)||
				 !strcmp(sortjobname,joblist[i].id)||
			 	 !strcmp(sortprogramname,joblist[i].jobname)
			 	){
					if(i==cursor&&cw==QUEUEWINDOW) {
						attron(A_BOLD);
						strcpy(job,joblist[i].id);
					}
	       	/*printw("%7s %5s  %10s  %10s  %4s    %10s %8s   %3i \n",*/
	    	  printw("%7s  %4s %10s  %10s %4s %10s %8s  %3i \n",
					joblist[i].id,joblist[i].priority,joblist[i].jobname,
					joblist[i].username,joblist[i].state,joblist[i].submittime,
					joblist[i].submitdate, joblist[i].nprocs);
					if(i==cursor&&cw==QUEUEWINDOW) attroff(A_BOLD);
				/*count the jobs printed */
				printed++;
				}
				else cursor++;
			}
			else {
				if(i==cursor&&cw==QUEUEWINDOW) {
					attron(A_BOLD);
					strcpy(job,joblist[i].id);
				}
	    	/*printw("%7s %5s  %10s  %10s  %4s    %10s %8s   %3i \n",*/
	    	printw("%7s  %4s %10s  %10s %4s %10s %8s  %3i \n",
					joblist[i].id,joblist[i].priority,joblist[i].jobname,
					joblist[i].username,joblist[i].state,joblist[i].submittime,
					joblist[i].submitdate, joblist[i].nprocs);
				if(i==cursor&&cw==QUEUEWINDOW) attroff(A_BOLD);
				printed++;
		}
	}
	jobsortcount=printed;
}
/* convert memory values to floats, check the size (K,M,G)
 * put everything in Megabyte format */
float mem_convert(char *value){
	float number;

	number=atof(value);
	if(number>0){
		switch(value[strlen(value)-1]){	
			case 'K':
				number=number/1024;
				break;
			case 'G':
				number=number*1024;
				break;
			case 'M':
				break;
			default:
				number=0;
				break;
		}
	}
	else number=0;
	return number;
}
int mem_unit(float *number){

	if(number>0){
		if(*number>=1024){
			*number=*number/1024;
			return 'G';
		}
		if(*number<1){
			*number=*number*1024;
			return 'K';
		}
		return 'M';
	}
	return ' ';
}

void do_text_error(char *error){
			endwin();
			printf("Error: %s\n",error);
			exit(1);
}
void display_top(){
  int maxy,maxx,i;
	char top_data[MAXTOPDATA];
/*FILE *fp; */


  clear();
  getmaxyx(stdscr,maxy,maxx);
  if(maxy<23||maxx<strlen(HOSTLABLE2)) printw("Window too small for top. Press a key");
  else{
		attron(A_REVERSE);
  	printw("Top Snapshot For Node: ");
		attron(A_BOLD);
		printw("%s  ",top_node);
		attroff(A_BOLD);
		printw("Press any key to exit.\n");
		attroff(A_REVERSE);
	  top_found=0;
		gettopdata(top_data,maxx-2,maxy-2);
		if (strlen(top_data) > 0) printw("%s",top_data); 
		else printw("\n Not top data avaiable");
	}
  noecho(); cbreak(); getch(); nocbreak(); echo();
}

void display_help(){
	int maxy,maxx;
	
	clear();
	getmaxyx(stdscr,maxy,maxx);
  if(maxy<23||maxx<strlen(HOSTLABLE2)) printw("Window too small for help. Press a key");
	else
	printw("Userstat displays node statistics and the batch queue on a cluster.\n\
The top window is the batch queue and the bottom window are the hosts. The\n\
windows update automatically and are scrollable with the arrow keys. A \"+\" \n\
indicates that the list will scroll further. Available options:\n\
  q - to quit userstat\n\
  h - to get this help\n\
  b - to make the batch window active\n\
  n - to make the nodes window active\n\
  spacebar - update windows (automatic update after 20 seconds)\n\
  up_arrow - to move though the jobs or nodes window\n\
  down_arrow - to move though the jobs or nodes window\n\
  Pg Up/Down - move a whole page in the jobs or nodes window\n\
\n");
attron(A_BOLD);
printw("Queue Window Commands:                  Host Window Commands:\n");
attroff(A_BOLD);
printw("  j - sort on job-ID                     s - sort on system hostname\n\
  u - sort on user name                  a - redisplay all hosts\n\
  p - sort on program name               return - top snapshot for node\n\
  a - redisplay all jobs                 *nodename* means node is down \n\
  d - delete a job from the queue\n\
  return - display only the nodes for that job\n\
  (When sorting on multiple parameters all matches are displayed.)\n\
\n");
attron(A_BOLD);
printw("Press any key to continue");
attroff(A_BOLD);

	noecho();cbreak();getch();nocbreak();echo();
}

/* determines is job index in already in job array 
 * returns -1 if not, the index if it is, 0 is a
 * valid index.
 */
int inarray(char *id,int index){
	int i=0;
	for(i=0;i<index;i++){
		if (!strcmp(id,joblist[i].id)){
			return i;
		}
	}
	return -1;
}

/* return the index of the hostlist for "hostname"
 * in the hostlist array.
 */
int gethostindex(char *hostname,int nodecount){
	int i;
	for (i=0;i<nodecount;i++){
		if (!strcmp(hostname,hostlist[i].hostname))
			break;
	}
	return i;
}

/* read the host data from the file an place in the hostdata array */
void gethostdata(){

	char line[MAXLINELEN],*token[STRINGLEN];
	int i,j;
	int pipefds[2];
	pid_t pid;
	
	down=0;
	qhost_found=0;
	pipe(pipefds);
	pid=pipe2qhost(pipefds[FD_WRITE],pipefds[FD_READ]);
  while (pipefgets(line,MAXLINELEN-1,pipefds[FD_READ])!=NULL){    
		qhost_found=1; /* qstat returned response */
		if((token[0]=strtok(line," \n"))!=NULL){
			if (strncmp(token[0],SKIP1,6)!=0 &&
		  	  strncmp(token[0],SKIP2,6)!=0 &&
		    	strncmp(token[0],SKIP3,6)!=0 ){
				strcpy(hostlist[nodecount].hostname,token[0]);
				token[0]=strtok(NULL," \n");
				/*skip the ARCH */
				if((token[0]=strtok(NULL," \n"))!=NULL){
       		hostlist[nodecount].processors=atoi(token[0]);
					tprocessors+=hostlist[nodecount].processors;
				}
				if((token[0]=strtok(NULL," \n"))!=NULL){
					if(!strcmp("-",token[0])){ /*down node */
						down++;
						hostlist[nodecount].down=1;
					}
					else hostlist[nodecount].down=0;
       		hostlist[nodecount].load=atof(token[0]);
					tload+=hostlist[nodecount].load;
				}
				if((token[0]=strtok(NULL," \n"))!=NULL){
       		hostlist[nodecount].mem=mem_convert(token[0]);
					tmem+=hostlist[nodecount].mem;
				}
				if((token[0]=strtok(NULL," \n"))!=NULL){
       		hostlist[nodecount].memuse=mem_convert(token[0]);
					tmemuse+=hostlist[nodecount].memuse;
				}
				if((token[0]=strtok(NULL," \n"))!=NULL){
       		hostlist[nodecount].swap=mem_convert(token[0]);
					tswap+=hostlist[nodecount].swap;
				}
				if((token[0]=strtok(NULL," \n"))!=NULL){
						token[strlen(token[0])]=0;
       		hostlist[nodecount].swapuse=mem_convert(token[0]);
					tswapuse+=hostlist[nodecount].swapuse;
				}
				nodecount++;
			}
		}

 }
#if CLDEBUG
    printf("SUMMARY:\nnodes=%i tprocs=%i tload=%3.2f tmem=%3.1f tmemuse=%3.1f tswap=%3.1f tswapuse=%3.1f\n",nodecount,tprocessors,tload,tmem,tmemuse,tswap,tswapuse);
		puts("NODES:");
    for (i=0;i<nodecount;i++){
	    printf("%i %12s  %4i     %5.2f  %8.2f %8.2f %8.2f %8.2f",i,hostlist[i].hostname,hostlist[i].processors,hostlist[i].load,hostlist[i].mem,hostlist[i].memuse,hostlist[i].swap,hostlist[i].swapuse);
		}
#endif
	close(pipefds[FD_READ]);
 	waitpid(pid, NULL, 0);
}
/* reads and parses the output of qstat. 
 * It will first attmpt to read the file, if the fiel does not exist, it will
 * run qstat. If 
 */
void getqueuedata(){


	int i,j, blank=0,skip=0,master,masterisfirst;	
	char line[MAXLINELEN],*token[STRINGLEN], previous[STRINGLEN],hostname[STRINGLEN],hostname1[STRINGLEN];
	int pipefds[2];
	pid_t pid;


	pipe(pipefds);
	pid=pipe2qstat(pipefds[FD_WRITE],pipefds[FD_READ]);
	while (pipefgets(line,MAXLINELEN-1,pipefds[FD_READ])!=NULL){    
		master=0;
		if (!strncmp(line,"       ",7)){
			/* this is a SLAVE program spawned by MASTER program on same node use the
			 * previous job id for a second copy of the program on the same node*/
			skip++;
			blank=1;
			if(skip>=2) masterisfirst=0;
		}
		else {masterisfirst=0; /*this is the case where the MASTEr line was the first line
														encountered for this job*/
						skip=0;
		}
		 if(strstr(line,"MASTER")){
			master=1; /* if this is the master line we may ignore it for parallel jobs*/
		}	
		if((token[0]=strtok(line," "))!=NULL){
			if (strncmp(token[0],SKIP4,6)!=0 &&
		    strncmp(SKIP2, token[0],6)!=0 ){
		    /* if in array already, return the index where it was found,
		     * then increment procs and add the proc index from 
		     * hostlist to the nodelist.
		     * If it is not in the array, then add it.*/
				if(!skip){
					jobindex=inarray(token[0],jobcount);
			 		 strcpy(previous,token[0]);
					}
				else{
					jobindex=inarray(previous,jobcount);
				}
		    if(jobindex>=0) {
					 if(!master){ /* ignore the master line*/
						if(!masterisfirst) (joblist[jobindex].nprocs)++;
          	/* determine the host name (using cluster@nodename syntax)
           	 *put the index to the host in the hostlist 
						 * skip all other fields
						 */
						if(blank){
		        	strcpy(hostname1,token[0]);
							i=0;j=0;
							while(hostname1[i]!= '@' ){i++;}
							i++;
							while(hostname1[i+j]!=0){hostname[j]=hostname1[i+j];j++;}
							hostname[j]=0;
							blank=0;
          		joblist[jobindex].nodelist[joblist[jobindex].nprocs]=gethostindex(hostname,nodecount);
						}
						else{

						if(!skip) token[0]=strtok(NULL," ");
						token[0]=strtok(NULL," ");
						token[0]=strtok(NULL," ");
						if((token[0]=strtok(NULL," "))!=NULL){
							if(!strcmp(joblist[jobindex].state,"r") /* only r and t */
								||!strcmp(joblist[jobindex].state,"t")
								||!strcmp(joblist[jobindex].state,"Rt")
								||!strcmp(joblist[jobindex].state,"dr")
								||!strcmp(joblist[jobindex].state,"dt")){
								token[0]=strtok(NULL," ");
								token[0]=strtok(NULL," ");
								if((token[0]=strtok(NULL," "))!=NULL){
		        			strcpy(hostname1,token[0]);
									i=0;j=0;
									while(hostname1[i]!= '@' ){i++;}
									i++;
									while(hostname1[i+j]!=0){hostname[j]=hostname1[i+j];j++;}
									hostname[j]=0;
								}
          			joblist[jobindex].nodelist[joblist[jobindex].nprocs]=gethostindex(hostname,nodecount);
							}
						else (joblist[jobindex].nprocs)--; /* this was Rq or Rt-ignore it*/
					}
				}	
					/*else (joblist[jobindex].nprocs)--; /* because we count the LINE twice*/
				}
			 /*else strcpy(previous,token[0]);*/
		    }	    
		    else{
					jobindex=jobcount;
					joblist[jobindex].nprocs=0;
					if (!skip){	/* no blank, so do not skip, copy the token to id and previous*/
						strcpy(joblist[jobindex].id,token[0]);
						strcpy(previous,token[0]);
					}
					else{	/* this is a skip, use the previous job id as the id */
						strcpy(joblist[jobindex].id,previous);
					}
					if(!skip){	/* no blank, read the next value */
						if((token[0]=strtok(NULL," "))!=NULL)
		        	strcpy(joblist[jobindex].priority,token[0]);
					}else	/* this is skip, existing token is the priorty */
						strcpy(joblist[jobindex].priority,token[0]);
					if((token[0]=strtok(NULL," "))!=NULL)
		       	strcpy(joblist[jobindex].jobname,token[0]);
					if((token[0]=strtok(NULL," "))!=NULL)
		       	strcpy(joblist[jobindex].username,token[0]);
					if((token[0]=strtok(NULL," "))!=NULL)
		       	strcpy(joblist[jobindex].state,token[0]);
					if((token[0]=strtok(NULL," "))!=NULL)
		       	strcpy(joblist[jobindex].submitdate,token[0]);
					if((token[0]=strtok(NULL," "))!=NULL)
		       	strcpy(joblist[jobindex].submittime,token[0]);
					if(!strcmp(joblist[jobindex].state,"r")
						||!strcmp(joblist[jobindex].state,"t")
						||!strcmp(joblist[jobindex].state,"Rt")
						||!strcmp(joblist[jobindex].state,"dt")
						||!strcmp(joblist[jobindex].state,"dr")){ /* only t and r jobs have cpus */
						if((token[0]=strtok(NULL," "))!=NULL){
		      
		        			strcpy(hostname1,token[0]);
									i=0;j=0;
									while(hostname1[i]!= '@' ){i++;}
									i++;
									while(hostname1[i+j]!=0){hostname[j]=hostname1[i+j];j++;}
									hostname[j]=0;
						}
						if((token[0]=strtok(NULL," "))!=NULL){
		        	if(!strcmp("MASTER",token[0]))
								masterisfirst=1;
							else
								masterisfirst=0;
							joblist[jobindex].nprocs=1;
							/* determine the host name (using hostname.q syntax)
      	 			 * put the index to the host in the hostlist */
							joblist[jobindex].nodelist[joblist[jobindex].nprocs]=gethostindex(hostname,nodecount);
				
						}
					/*joblist[jobindex].nprocs=0;  every ting else has zero cpus associated with it */
					}
				jobcount++;
	  		}	
	 		}
		}
 	}
 #if CLDEBUG
		printf("JOB QUEUE - total jobs=%i:\n",jobcount);
    for (i=0;i<jobcount;i++){
	    printf("%i %s %s %s %s %s %s %s %i \n",i,joblist[i].id,joblist[i].priority,joblist[i].jobname,joblist[i].username,joblist[i].state,joblist[i].submittime,joblist[i].submitdate, joblist[i].nprocs);
		}
		puts("NODES USED:");
    for (i=0;i<jobcount;i++){
			printf("%i id=%s nprocs=%i\n",i,joblist[i].id,joblist[i].nprocs);
				for(j=1;j<=joblist[i].nprocs;j++){
						printf("     node=%s \n",hostlist[joblist[i].nodelist[j]].hostname);
				}
		}
 #endif 
  close(pipefds[FD_READ]);
 	waitpid(pid, NULL, 0);
}

void gettopdata(char *top_data,int max_line_len, int max_lines){

  int i,lines=0;
	char top_line[MAXLINELEN];
  int pipefds[2];
  pid_t pid;

	top_data[0]=0;
 	pipe(pipefds);
	pid=pipe2top(pipefds[FD_WRITE],pipefds[FD_READ]);
 	while ((pipefgets(top_line,MAXLINELEN-1,pipefds[FD_READ])!=NULL)) {
		if (lines<max_lines){
   	strncat(top_data,top_line,max_line_len);
		if (top_data[strlen(top_data)-1] != '\n')
			top_data[strlen(top_data)-1]='\n';
		lines++;
		}
		i=strlen(top_data);
	}	
 	close(pipefds[FD_READ]);
 	waitpid(pid, NULL, 0);
}

pid_t pipe2qhost(int fdw, int fdr){
		pid_t pid;

	pid = fork();
	if (pid == 0) {
		/*close(fdr);*/
		dup2(fdw,1);
		if (execlp("qhost","qhost", NULL)){
			close(fdw);
			endwin();
      perror("qhost");
      exit(1);
 		 }
	}else if (pid != -1) {
			close(fdw);
			return pid;
	}
 	else 
    	 printf("There was an error with forking.\n");
}
pid_t  pipe2qstat(int fdw, int fdr){
		pid_t pid;

	pid = fork();
	if (pid == 0) {
		dup2(fdw,1);
		 if (execlp("qstat","qstat", "-g","t","-u","*",NULL)){ 
			close(fdw);
			endwin();
      perror("qstat");
      exit(1);
 		 }
	}else if (pid != -1) {
			close(fdw);
			return pid;
	}
 	 else 
    	 printf("There was an error with forking.\n");
}
pid_t  pipe2top(int fdw, int fdr){
		pid_t pid;

	pid = fork();
	if (pid == 0) {
		dup2(fdw,1);
		 if (execlp("ssh","ssh",top_node,"top","-b","-n1",NULL)){  
			close(fdw);
			endwin();
      perror("ssh_top");
      exit(1);
 		 }
	}else if (pid != -1) {
			close(fdw);
			return pid;
	}
 	 else 
    	 printf("There was an error with forking.\n");
}


char *pipefgets(char *line,size_t max,int fd){
	size_t count=0,ret;
	void *pts = line;
	
	while((ret=read(fd,pts+count,1))!=0){
		if(count>max)
				return NULL;
		if(line[count]=='\n'){
			line[count+1]=0;
			return line;
		}
		count++;
	}
  return NULL;
}
void delete_job(char *jobname){
    pid_t pid;

	 pid = fork();
	 if (pid == 0) {
			if (execlp("qdel","qdel", jobname,NULL)){
		    endwin();
		    perror("qdel");
		    exit(1);
		  }
	}else if (pid != -1) {
	  waitpid(pid, NULL, 0);
	}
	else 
	  printf("There was an error with forking.\n");
}
